/*
 * Copyright (c) 1996, 1997 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <Dem/usgsdem.h>

#ifndef MAKEDEPEND
#include <vtkDecimatePro.h>
#include <vtkFloatPoints.h>
#include <vtkFloatScalars.h>
#include <vtkPolyDataWriter.h>
#include <vtkStructuredGrid.h>
#include <vtkStructuredGridGeometryFilter.h>
#include <vtkTriangleFilter.h>
#endif

extern "C" {
#include <projects.h>
#include <sys/stat.h>
#include <unistd.h>
}

#include <stdio.h>
#include <string.h>

short round(float x) {
    return x > 0 ? short(x+0.5) : -short(-x+0.5);
}

int main(int argc, char* argv[]) {
    if (argc != 3) {
	printf("usage: usgsdem2vtk dem targetdir\n");
	return(-1);
    }

    UsgsDem* dem = new UsgsDem();
    dem->read(argv[1]);

    if (dem->reference_system() == UsgsDem::GEOGRAPHIC &&
	dem->rows() == 1201 && dem->cols() == 1201) {
	// 1 degree USGS DEM
	for (int k = 0; k < 6; k++) {
	    for (int l = 0; l < 6; l++) {
		vtkStructuredGrid* grid = vtkStructuredGrid::New();
		grid->SetDimensions(201, 201, 1);
		vtkFloatPoints* points = vtkFloatPoints::New();
		points->Allocate(201 * 201);
		double dx, dy;
		float p[3];
		for (int j = 0; j < 201; j++)
		    for (int i = 0; i < 201; i++) {
			int di = k * 200 + i;
			int dj = l * 200 + j;
			dem->projected_position(dj, di, dx, dy);
			p[0] = dx;
			p[1] = dy;
			p[2] = dem->value(dj, di);
			points->InsertNextPoint(p);
		    }
		grid->SetPoints(points);
		points->Delete();
		
		vtkStructuredGridGeometryFilter* gfilter = 
		    vtkStructuredGridGeometryFilter::New();
		gfilter->SetExtent(0, 200, 0, 200, 0, 0);
		gfilter->SetInput(grid);
		
		vtkTriangleFilter* tfilter = vtkTriangleFilter::New();
		tfilter->PassVertsOff();
		tfilter->PassLinesOff();
		tfilter->SetInput(gfilter->GetOutput());
		
		vtkDecimatePro* dfilter = vtkDecimatePro::New();
		dfilter->SetTargetReduction(0.95);
		dfilter->PreserveTopologyOff();
		dfilter->SplittingOff();
		dfilter->BoundaryVertexDeletionOff();
		dfilter->SetMaximumError(10.0);
		dfilter->SetInput(tfilter->GetOutput());
		dfilter->Update();
		
		vtkPolyData* decdata = dfilter->GetOutput();
		float bounds[6];
		decdata->GetBounds(bounds);
		char bdattrs[256];
		sprintf(bdattrs, "%.4f %.4f %.4f %.4f %.4f %.4f",
			bounds[0], bounds[1], bounds[2], bounds[3],
			bounds[4], bounds[5]);

		int npt = decdata->GetNumberOfPoints();
		vtkFloatScalars* sc = vtkFloatScalars::New();
		sc->Allocate(npt);
		float point[3];
		for (int pti = 0; pti < npt; pti++) {
		  decdata->GetPoint(pti, point);
		  sc->InsertNextScalar(point[2]);
		}
		decdata->GetPointData()->SetScalars(sc);
		sc->Delete();

		int flen = strlen(argv[2]) + 80;
		char* fname = new char[flen];
		memset(fname, 0, flen);
		strcpy(fname, argv[2]);
		if (fname[strlen(fname)-1] != '/')
		    strcat(fname, "/");
		const char* demname = dem->Name();
		char c;
		while (c = *demname++)
		    if (!isspace(c))
			fname[strlen(fname)] = c;
		fname[strlen(fname)] = '.';
		fname[strlen(fname)] = '0' + k;
		fname[strlen(fname)] = '.';
		fname[strlen(fname)] = '0' + l;
		strcat(fname, ".vtk");
		printf("filename is %s\n", fname);
#if 0
		vtkPolyDataWriter* writer = vtkPolyDataWriter::New();
		writer->SetFileName(fname);
		writer->SetHeader(fname);
		writer->SetFileType(VTK_ASCII);
		writer->SetInput(decdata);
		writer->Write();
#endif
		FILE* fp = fopen(fname, "w");
		if (fp) {
		  fprintf(fp, "# vtk DataFile Version 2.0\n");
		  fprintf(fp, "%s\n", bdattrs);
		  fprintf(fp, "ASCII\nDATASET POLYDATA\n");
		  fprintf(fp, "POINTS %d float\n", npt);
		  float point[3];
		  for (int pti = 0; pti < npt; pti++) {
		    decdata->GetPoint(pti, point);
		    fprintf(fp, "%f %f %f ", point[0], point[1], point[2]);
		  }
		  int ncells = decdata->GetPolys()->GetNumberOfCells();
		  int size = decdata->GetPolys()->GetNumberOfConnectivityEntries();
		  fprintf(fp, "\nPOLYGONS %d %d\n", ncells, size);
		  int npts, *pts, pj;
		  for (decdata->GetPolys()->InitTraversal();
		       decdata->GetPolys()->GetNextCell(npts, pts); ) {
		    fprintf(fp, "%d ", npts);
		    for (pj=0; pj<npts; pj++)
		      fprintf(fp, "%d ", pts[pj]);
		    fprintf(fp, "\n");
		  }
		  fprintf(fp, "\nPOINT_DATA %d\n", npt);
		  fprintf(fp, "SCALARS scalars float\n");
		  fprintf(fp, "LOOKUP_TABLE default\n");
		  for (int si = 0; si < npt; si++) {
		    float s = decdata->GetPointData()->GetScalars()->GetScalar(si);
		    fprintf(fp, "%f ", s);
		    if (!((si+1)%6)) fprintf(fp, "\n");
		  }
		  fprintf(fp, "\n");
		  fclose(fp);
		}
		else {
		  fprintf(stderr, "could not open %s\n", fname);
		}

#if 0
	// this creates a geom file for lib3d.  format is similar
	// to vtk except vertex ids are 1-offset
	// also translating to center near origin to facilitate
	// initial lib3d tests with the testobj program
	char gfname[256];
	strcpy(gfname, fname);
	strcat(gfname, ".geom");
	FILE* fp = fopen(gfname, "w");

	int np = decdata->GetNumberOfPoints();
	int nc = decdata->GetNumberOfCells();
	fprintf(fp, "%d\t%d\t%d\t\n", np, nc, nc);
	for (int i = 0; i < np; i++) {
	  float pt[3];
	  decdata->GetPoints()->GetPoint(i, pt);
	  //addVertex(pt[0], pt[1], pt[2]);
	  fprintf(fp, "%.6f\t%.6f\t%.6f\t\n",
		  pt[0]-(bounds[0]+(bounds[1]-bounds[0])/2.0),
		  pt[2]-bounds[5]-20.0,
		  pt[1]-(bounds[2]+(bounds[3]-bounds[2])/2.0));
	}

	for (int i = 0; i < nc; i++) {
	  vtkIdList ptIds;
	  decdata->GetCellPoints(i, ptIds);
	  //addTriangle(ptIds.GetId(0), ptIds.GetId(1), ptIds.GetId(2));
	  fprintf(fp, "3\t%d\t%d\t%d\t\n", ptIds.GetId(0)+1, ptIds.GetId(1)+1, ptIds.GetId(2)+1);
	}
	fclose(fp);
#endif
		
	//writer->Delete();
		//dfilter->Delete();
		tfilter->Delete();
		gfilter->Delete();
		grid->Delete();
	    }
	}
    }
    else if (dem->reference_system() == UsgsDem::UTM) {
#if 0
      long xmin, ymin, xmax, ymax;
      dem->utm_bbox(xmin, ymin, xmax, ymax);
      int skip, level;
      for (int ri = 0; ri < 4; ri++) {
	for (int rj = 0; rj < 4; rj++) {
	  for (skip = 1, level = 0; skip < 64; skip *= 2, level++) {
	    vtkPolyData* polydata = vtkPolyData::New();
	    vtkCellArray* cells = vtkCellArray::New();
	    vtkIntPoints* intpoints = NULL;
	    vtkFloatPoints* fltpoints = NULL;
	    vtkShortScalars* shortscalars = NULL;
	    vtkFloatScalars* floatscalars = NULL;
	    if (dem->elevation_units() == UsgsDem::METERS) {
	      intpoints = vtkIntPoints::New();
	      shortscalars = vtkShortScalars::New();
	    }
	    else if (dem->elevation_units() == UsgsDem::FEET) {
	      fltpoints = vtkFloatPoints::New();
	      floatscalars = vtkFloatScalars::New();
	    }
	    else
	      printf("warning: elevation not in meter or feet units\n");
	    int intp[3];
	    float fltp[3];

	    int sskip = 1;
	    int rows = (64 / sskip) + 1;
	    int cols = (64 / sskip) + 1;
	    int* ptidx = new int[rows * cols];
	    for (int n = 0; n < rows * cols; n++)
	      ptidx[n] = -1;
	    int idx = 0;
	    int startrow =  64 * ri;
	    int startcol = 110 + 64 * rj;
	    int i, j, ii, jj;
	    for (i = startcol, ii = 0; i <= startcol + 64; i += sskip, ii++) {
	      for (j = startrow, jj = 0; j <= startrow + 64; j += sskip, jj++) {
	        if (dem->elevation_units() == UsgsDem::METERS) {
		  intp[0] = xmin + i * 30;
		  intp[1] = ymin + j * 30;
		  intp[2] = dem->value(j, i);
		  intpoints->InsertNextPoint(intp);
		  shortscalars->InsertNextScalar((short)dem->value(j, i));
		}
		else if (dem->elevation_units() == UsgsDem::FEET) {
		  fltp[0] = xmin + i * 30;
		  fltp[1] = ymin + j * 30;
		  fltp[2] = dem->value(j, i) * 1.0/3.28;
		  fltpoints->InsertNextPoint(fltp);
		  floatscalars->InsertNextScalar(dem->value(j, i) * 1.0/3.28);
		}
		ptidx[jj + rows * ii] = idx++;
	      }
	    }
	    for (i = 0; i < cols - 1; i++) {
	      for (j = 0; j < rows - 1; j++) {
	        if (ptidx[(j+1) + rows * i] > 0 &&
		    ptidx[(j+1) + rows * (i+1)] > 0) {
		  intp[0] = ptidx[j + rows * i];
		  intp[1] = ptidx[(j+1) + rows * i];
		  intp[2] = ptidx[(j+1) + rows * (i+1)];
		  cells->InsertNextCell(3, intp);
		}
	        if (ptidx[j + rows * (i+1)] > 0 &&
		    ptidx[(j+1) + rows * (i+1)] > 0) {
		  intp[0] = ptidx[j + rows * i];
		  intp[1] = ptidx[j + rows * (i+1)];
		  intp[2] = ptidx[(j+1) + rows * (i+1)];
		  cells->InsertNextCell(3, intp);
		}
	      }
	    }
	    //	    delete [] ptidx;
	    if (dem->elevation_units() == UsgsDem::METERS) {
	      polydata->SetPoints(intpoints);
	      intpoints->Delete();
	    }
	    else if (dem->elevation_units() == UsgsDem::FEET) {
	      polydata->SetPoints(fltpoints);
	      fltpoints->Delete();
	    }
	    polydata->SetPolys(cells);
	    cells->Delete();
	    if (dem->elevation_units() == UsgsDem::METERS) {
	      polydata->GetPointData()->SetScalars(shortscalars);
	      shortscalars->Delete();
	    }
	    else if (dem->elevation_units() == UsgsDem::FEET) {
	      polydata->GetPointData()->SetScalars(floatscalars);
	      floatscalars->Delete();
	    }

	    vtkPolyData* dataset = polydata;
	    //if (level == 0) {
	      Decimator* dfilter = new Decimator();
	      //dfilter->DebugOn();
	      //dfilter->SetErrorIncrement(0.07);
	      //dfilter->SetMaximumError(0.02);
	      //dfilter->SetFeatureAngleIncrement(1.0);
	      dfilter->SetTargetReduction(0.1 * (level+1));
	      dfilter->SetMaximumIterations(1);
	      //dfilter->SetMaximumSubIterations(1);
	      dfilter->SetErrorIncrement(0.03);
	      dfilter->SetMaximumError(0.05);
	      dfilter->PreserveEdgesOn();
	      dfilter->BoundaryVertexDeletionOff();
	      dfilter->SetInput(polydata);
	      dfilter->Update();
	      dataset = dfilter->GetOutput();
	      //}

	    int flen = strlen(argv[2]) + 80;
	    char* fname = new char[flen];
	    memset(fname, 0, flen);
	    strcpy(fname, argv[2]);
	    if (fname[strlen(fname)-1] != '/')
	      strcat(fname, "/");
	    const char* demname = dem->Name();
	    char c;
	    while (c = *demname++)
	      if (!isspace(c))
		fname[strlen(fname)] = c;
	    fname[strlen(fname)] = '.';
	    fname[strlen(fname)] = '0' + ri;
	    fname[strlen(fname)] = '.';
	    fname[strlen(fname)] = '0' + rj;
	    fname[strlen(fname)] = '.';
	    fname[strlen(fname)] = '0' + level;
	    strcat(fname, ".vtk");
	    printf("filename is %s\n", fname);
	    vtkPolyWriter* writer = new vtkPolyWriter();
	    writer->SetFilename(fname);
	    writer->SetHeader(fname);
	    writer->SetFileType(VTK_ASCII);
	    writer->SetInput(dataset);
	    writer->Write();
	    writer->Delete();
	    polydata->Delete();

	  }
	}
      }
#endif
#if 1
        vtkPolyData* polydata = vtkPolyData::New();
	vtkCellArray* cells = vtkCellArray::New();
	vtkFloatPoints* fltpoints = NULL;
	fltpoints = vtkFloatPoints::New();
	float fltp[3];
	long xmin, ymin, xmax, ymax;
	dem->utm_bbox(xmin, ymin, xmax, ymax);
	int* ptidx = new int[dem->rows() * dem->cols()];
	for (int n = 0; n < dem->rows() * dem->cols(); n++)
	    ptidx[n] = -1;
	int idx = 0;
	for (int i = 0; i < dem->cols(); i++) {
	  //for (int j = dem->start_row(i); j <= dem->end_row(i); j++) {
	    for (int j = 0; j <= dem->rows(); j++) {
	        if (dem->elevation_units() == UsgsDem::METERS) {
	            fltp[0] = xmin + i * 30;
		    fltp[1] = ymin + j * 30;
		    fltp[2] = dem->value(j, i);
		    fltpoints->InsertNextPoint(fltp);
		}
		else if (dem->elevation_units() == UsgsDem::FEET) {
		    fltp[0] = xmin + i * 30;
		    fltp[1] = ymin + j * 30;
		    fltp[2] = dem->value(j, i) * 1.0/3.28;
		    fltpoints->InsertNextPoint(fltp);
		}
		ptidx[j + dem->rows() * i] = idx++;
	    }
	}
	int intp[3];
	for (int i = 0; i < dem->cols() - 1; i++) {
	    for (int j = dem->start_row(i); j < dem->end_row(i); j++) {
	        if (ptidx[(j+1) + dem->rows() * i] > 0 &&
		    ptidx[(j+1) + dem->rows() * (i+1)] > 0) {
		    intp[0] = ptidx[j + dem->rows() * i];
		    intp[1] = ptidx[(j+1) + dem->rows() * i];
		    intp[2] = ptidx[(j+1) + dem->rows() * (i+1)];
		    cells->InsertNextCell(3, intp);
		}
	        if (ptidx[j + dem->rows() * (i+1)] > 0 &&
		    ptidx[(j+1) + dem->rows() * (i+1)] > 0) {
		    intp[0] = ptidx[j + dem->rows() * i];
		    intp[1] = ptidx[j + dem->rows() * (i+1)];
		    intp[2] = ptidx[(j+1) + dem->rows() * (i+1)];
		    cells->InsertNextCell(3, intp);
		}
	    }
	}
	delete [] ptidx;
	polydata->SetPoints(fltpoints);
	fltpoints->Delete();

	polydata->SetPolys(cells);
	cells->Delete();

	//float bds[6];
	//polydata->GetBounds(bds);
	//printf("bounds: %f %f %f %f %f %f\n",
	//bds[0], bds[1], bds[2], bds[3], bds[4], bds[5]);
#if 0
	return 0;
	vtkPolyWriter* writer0 = vtkPolyWriter::New();
	writer0->SetFilename("sfsdem.xyz");
	writer0->SetHeader("sfsdem.xyz");
	writer0->SetFileType(VTK_ASCII);
	writer0->SetInput(polydata);
	writer0->Write();
	return 0;
#endif
	vtkDecimatePro* dfilter = vtkDecimatePro::New();
	dfilter->SetTargetReduction(0.95);
	dfilter->PreserveTopologyOff();
	dfilter->SplittingOff();
	dfilter->BoundaryVertexDeletionOff();
	dfilter->SetMaximumError(10.0);
	dfilter->SetInput(polydata);
	dfilter->Update();
	vtkPolyData* decdata = dfilter->GetOutput();

	int npt = decdata->GetNumberOfPoints();
	vtkFloatScalars* sc = vtkFloatScalars::New();
	sc->Allocate(npt);
	float point[3];
	for (int pti = 0; pti < npt; pti++) {
	  decdata->GetPoint(pti, point);
	  sc->InsertNextScalar(point[2]);
	}
	decdata->GetPointData()->SetScalars(sc);
	sc->Delete();

	float bounds[6];
	decdata->GetBounds(bounds);
	char bdattrs[256];
	sprintf(bdattrs, "%.4f %.4f %.4f %.4f %.4f %.4f",
		bounds[0], bounds[1], bounds[2], bounds[3],
		bounds[4], bounds[5]);

	int flen = strlen(argv[2]) + 80;
	char* fname = new char[flen];
	memset(fname, 0, flen);
	strcpy(fname, argv[2]);
	if (fname[strlen(fname)-1] != '/')
	  strcat(fname, "/");
	const char* demname = dem->Name();
	char c;
	while (c = *demname++)
	  if (!isspace(c))
	    fname[strlen(fname)] = c;
	strcat(fname, ".vtk");
	printf("filename is %s\n", fname);
#if 0
	vtkPolyDataWriter* writer = vtkPolyDataWriter::New();
	writer->SetFileName(fname);
	writer->SetHeader(fname);
	writer->SetFileType(VTK_ASCII);
	writer->SetInput(decdata);
	writer->Write();
#endif
	FILE* fp = fopen(fname, "w");
	if (fp) {
	  fprintf(fp, "# vtk DataFile Version 2.0\n");
	  fprintf(fp, "%s\n", bdattrs);
	  fprintf(fp, "ASCII\nDATASET POLYDATA\n");
	  fprintf(fp, "POINTS %d float\n", npt);
	  float point[3];
	  for (int pti = 0; pti < npt; pti++) {
	    decdata->GetPoint(pti, point);
	    fprintf(fp, "%f %f %f ", point[0], point[1], point[2]);
	  }
	  int ncells = decdata->GetPolys()->GetNumberOfCells();
	  int size = decdata->GetPolys()->GetNumberOfConnectivityEntries();
	  fprintf(fp, "\nPOLYGONS %d %d\n", ncells, size);
	  int npts, *pts, pj;
	  for (decdata->GetPolys()->InitTraversal();
	       decdata->GetPolys()->GetNextCell(npts, pts); ) {
	    fprintf(fp, "%d ", npts);
	    for (pj=0; pj<npts; pj++)
	      fprintf(fp, "%d ", pts[pj]);
	    fprintf(fp, "\n");
	  }
	  fprintf(fp, "\nPOINT_DATA %d\n", npt);
	  fprintf(fp, "SCALARS scalars float\n");
	  fprintf(fp, "LOOKUP_TABLE default\n");
	  for (int si = 0; si < npt; si++) {
	    float s = decdata->GetPointData()->GetScalars()->GetScalar(si);
	    fprintf(fp, "%f ", s);
	    if (!((si+1)%6)) fprintf(fp, "\n");
	  }
	  fprintf(fp, "\n");
	  fclose(fp);
	}
	else {
	  fprintf(stderr, "could not open %s\n", fname);
	}

#if 0
	// this creates a geom file for lib3d.  format is similar
	// to vtk except vertex ids are 1-offset
	// also translating to center near origin to facilitate
	// initial lib3d tests with the testobj program
	char gfname[256];
	strcpy(gfname, fname);
	strcat(gfname, ".geom");
	FILE* fp = fopen(gfname, "w");

	int np = decdata->GetNumberOfPoints();
	int nc = decdata->GetNumberOfCells();
	fprintf(fp, "%d\t%d\t%d\t\n", np, nc, nc);
	for (int i = 0; i < np; i++) {
	  float pt[3];
	  decdata->GetPoints()->GetPoint(i, pt);
	  //addVertex(pt[0], pt[1], pt[2]);
	  fprintf(fp, "%.6f\t%.6f\t%.6f\t\n",
		  pt[0]-(bds[0]+(bds[1]-bds[0])/2.0),
		  pt[2]-bds[5]-20.0,
		  pt[1]-(bds[2]+(bds[3]-bds[2])/2.0));
	}

	for (int i = 0; i < nc; i++) {
	  vtkIdList ptIds;
	  decdata->GetCellPoints(i, ptIds);
	  //addTriangle(ptIds.GetId(0), ptIds.GetId(1), ptIds.GetId(2));
	  fprintf(fp, "3\t%d\t%d\t%d\t\n", ptIds.GetId(0)+1, ptIds.GetId(1)+1, ptIds.GetId(2)+1);
	}
	fclose(fp);
#endif
	
	//writer->Delete();
	//dfilter->Delete();
	polydata->Delete();
#endif
    }
    else {
	printf("sorry, can only handle 7.5 minute or 1 degree USGS DEMs in this version.\n");
    }
}
